"""
    register.py - this module is a part of clb (cloud load balancer)

    Copyright 2010, Canonical Ltd.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from twisted.web.resource import Resource, ErrorPage
from Cheetah.Template import Template
import os
import tempfile
import sqlite3

serverdb = "clb.db"
DEFAULT_ID = '__DEFAULT__ID__'

class RegisterResource(Resource):

    def __init__(self, db, templatedir, outfile):
        self.db = db
        self.templatedir = templatedir
        self.outfile = outfile
        Resource.__init__(self)

    def getChild(self, name, request):
        # verify auth
        if request.getPassword() != get_preshared_key(self.db):
            request.responseHeaders.addRawHeader('www-authenticate','Basic realm="CLB Key"')
            return ErrorPage(401,'Authorization Required', '')
        if name == 'addserver':
            return AddServerResource(self.db, self.templatedir, self.outfile)
        if name == 'delserver':
            return DelServerResource(self.db, self.templatedir, self.outfile)
        return self

    def render_GET(self, request):
        return request.URLPath()

class AbstractServerResource(RegisterResource):
    def render_GET(self, request):
        return 'OK '+str(self.dparts)+self.messages

    def getChild(self, name, request):
        """ http://$who/addserver/+service/$what/$ip/$where """
        self.messages='--'
        parts = request.URLPath().pathList()
        self.dparts = dict()
        storepart = None
        for part in parts:
            if storepart and len(storepart):
                self.dparts[storepart]=part
                storepart = None
                continue
            if len(part) and part[0:3] == '%2B':
                storepart=part[3:]
        # store
        if name == '+save':
            # resolve special values
            if self.dparts['ip'] == u'%3Asource':
                self.dparts['ip'] = request.getClientIP()
            self.save()
        return self

    def updateConfigs(self):
        allservs = all_services(self.db)
        t = Template(file=os.path.join(self.templatedir,'haproxy.tmpl'), searchList=[allservs])
        outdir = os.path.dirname(self.outfile)
        ofile = tempfile.NamedTemporaryFile(dir=outdir,prefix="."+self.outfile)
        ofile.write(str(t))
        ofile.flush()
        try:
            os.unlink(self.outfile)
        except OSError:
            pass
        os.rename(ofile.name, self.outfile)
        try:
            ofile.close()
        except OSError:
            pass
        # TODO make it HUP haproxy here

class AddServerResource(AbstractServerResource):
    def save(self):
        cur = self.db.cursor()
        try:
            cur.execute('INSERT INTO servers (name,service,ip,port) VALUES (?,?,?,?)', (self.dparts['server'], self.dparts['service'], self.dparts['ip'], self.dparts['port'])) 
            self.messages += '[saved]'
            self.db.commit()
            self.updateConfigs()
        except sqlite3.Error, e:
            self.messages += '[sqlite_error=' + e.args[0] + ']'
            self.db.rollback()

class DelServerResource(AbstractServerResource):
    def save(self):
        cur = self.db.cursor()
        try:
            cur.execute('DELETE FROM servers WHERE name=? AND service=?', (self.dparts['server'], self.dparts['service']))
            self.messages += '[deleted]'
            self.db.commit()
            self.updateConfigs()
        except sqlite3.Error, e:
            self.messages += '[sqlite_error=' + e.args[0] + ']'
            self.db.rollback()

def set_preshared_key(db, key, id=None):
    cur = db.cursor()
    cur.execute('CREATE TABLE IF NOT EXISTS keys (id varchar(255), key varchar(255), primary key (id))')
    if id is None:
        id = DEFAULT_ID
    cur.execute("INSERT OR REPLACE INTO keys (key,id) VALUES (?,?)", (key, id))
    db.commit()

def get_preshared_key(db, id=None):
    if id is None:
        id = DEFAULT_ID
    cur = db.cursor()
    cur.execute("SELECT key FROM keys WHERE id=? LIMIT 1", ([id]))
    row = cur.fetchone()
    return row[0]

def all_services(db=None):
    if not db:
        db = sqlite3.connect(serverdb)

    db.row_factory = sqlite3.Row
    cur = db.cursor()
    cur.execute("SELECT COALESCE(si.ip, '0.0.0.0') as service_ip, COALESCE(si.port, '80') as service_port, srv.service as service,srv.name as name,srv.ip as ip, srv.port as port FROM servers srv LEFT OUTER JOIN services si ON si.name = srv.service ORDER BY srv.service")
    service = {'name':None}
    services = []
    servers = []
    rows = cur.fetchall()
    for row in rows:
        if service['name'] is None:
            service['name'] = row['service']
        if row['service'] != service['name']:
            service['servers'] = servers
            services.append(service)
            service = {}
            service['name'] = row['service']
            service['ip'] = row['service_ip']
            service['port'] = row['service_port']
            servers = []
        servers.append({'name':row['name'], 'ip':row['ip'],'port':row['port']})
    service = {}
    service['name'] = row['service']
    service['ip'] = row['service_ip']
    service['port'] = row['service_port']
    service['servers'] = servers
    services.append(service)
    return services
    
